@@ -1,7 +1,7 @@
 /* SASL server API implementation
  * Rob Siemborski
  * Tim Martin
- * $Id: client.c,v 1.36 2001/12/04 02:05:25 rjs3 Exp $
+ * $Id: client.c,v 1.50 2002/05/05 14:10:55 ken3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -60,6 +60,8 @@ static cmech_list_t *cmechlist; /* global var which holds the list */
 
 static sasl_global_callbacks_t global_callbacks;
 
+static int _sasl_client_active = 0;
+
 static int init_mechlist()
 {
   cmechlist->mutex = sasl_MUTEX_ALLOC();
@@ -98,6 +100,8 @@ static void client_done(void) {
   sasl_FREE(cmechlist);
 
   cmechlist = NULL;
+
+  _sasl_client_active = 0;
 }
 
 int sasl_client_add_plugin(const char *plugname,
@@ -201,7 +205,7 @@ int sasl_client_init(const sasl_callback_t *callbacks)
   if (ret!=SASL_OK)
     return ret;
 
-  sasl_client_add_plugin("EXTERNAL", &external_client_init);
+  sasl_client_add_plugin("EXTERNAL", &external_client_plug_init);
 
   ret = _sasl_common_init();
 
@@ -210,6 +214,11 @@ int sasl_client_init(const sasl_callback_t *callbacks)
 			       _sasl_find_getpath_callback(callbacks),
 			       _sasl_find_verifyfile_callback(callbacks));
   
+  if (ret == SASL_OK) {
+      _sasl_client_active = 1;
+      ret = _sasl_build_mechlist();
+  }
+
   return ret;
 }
 
@@ -268,6 +277,8 @@ int sasl_client_new(const char *service,
   int result;
   sasl_client_conn_t *conn;
   sasl_utils_t *utils;
+
+  if(_sasl_client_active==0) return SASL_NOTINIT;
   
   /* Remember, iplocalport and ipremoteport can be NULL and be valid! */
   if (!pconn || !service || !serverFQDN)
@@ -305,6 +316,7 @@ int sasl_client_new(const char *service,
   utils->conn= *pconn;
   conn->cparams->utils = utils;
   conn->cparams->canon_user = &_sasl_canon_user;
+  conn->cparams->flags = flags;
   
   result = _sasl_strdup(serverFQDN, &conn->serverFQDN, NULL);
   if(result == SASL_OK) return SASL_OK;
@@ -320,13 +332,13 @@ int sasl_client_new(const char *service,
 static int have_prompts(sasl_conn_t *conn,
 			const sasl_client_plug_t *mech)
 {
-  static const long default_prompts[] = {
+  static const unsigned long default_prompts[] = {
     SASL_CB_AUTHNAME,
     SASL_CB_PASS,
     SASL_CB_LIST_END
   };
 
-  const long *prompt;
+  const unsigned long *prompt;
   int (*pproc)();
   void *pcontext;
   int result;
@@ -382,6 +394,8 @@ int sasl_client_start(sasl_conn_t *conn,
     sasl_ssf_t bestssf = 0, minssf = 0;
     int result;
 
+    if(_sasl_client_active==0) return SASL_NOTINIT;
+
     if (!conn) return SASL_BADPARAM;
 
     /* verify parameters */
@@ -491,26 +505,22 @@ int sasl_client_start(sasl_conn_t *conn,
     c_conn->mech = bestm;
 
     /* init that plugin */
-    result = c_conn->mech->plug->mech_new(NULL,
+    result = c_conn->mech->plug->mech_new(c_conn->mech->plug->glob_context,
 					  c_conn->cparams,
 					  &(conn->context));
     if(result != SASL_OK) goto done;
 
     /* do a step -- but only if we can do a client-send-first */
  dostep:
-    if(c_conn->mech->plug->features & SASL_FEAT_INTERNAL_CLIENT_FIRST) {
-	/* The plugin handles client-first internally */
-	result = sasl_client_step(conn, NULL, 0, prompt_need,
-				  clientout, clientoutlen);
-    } else if(clientout) {
-	if(c_conn->mech->plug->features & SASL_FEAT_WANT_CLIENT_FIRST) {
-	    result = sasl_client_step(conn, NULL, 0, prompt_need,
-				      clientout, clientoutlen);
-	} else {
-	    *clientout = NULL;
-	    *clientoutlen = 0;
-	    result = SASL_CONTINUE;
-	}
+    if(clientout) {
+        if(c_conn->mech->plug->features & SASL_FEAT_SERVER_FIRST) {
+            *clientout = NULL;
+            *clientoutlen = 0;
+            result = SASL_CONTINUE;
+        } else {
+            result = sasl_client_step(conn, NULL, 0, prompt_need,
+                                      clientout, clientoutlen);
+        }
     }
     else
 	result = SASL_CONTINUE;
@@ -543,12 +553,22 @@ int sasl_client_step(sasl_conn_t *conn,
   sasl_client_conn_t *c_conn= (sasl_client_conn_t *) conn;
   int result;
 
+  if(_sasl_client_active==0) return SASL_NOTINIT;
   if(!conn) return SASL_BADPARAM;
 
   /* check parameters */
   if ((serverin==NULL) && (serverinlen>0))
       PARAMERROR(conn);
 
+  /* Don't do another step if the plugin told us that we're done */
+  if (conn->oparams.doneflag) {
+      _sasl_log(conn, SASL_LOG_ERR, "attempting client step after doneflag");
+      return SASL_FAIL;
+  }
+
+  if(clientout) *clientout = NULL;
+  if(clientoutlen) *clientoutlen = 0;
+
   /* do a step */
   result = c_conn->mech->plug->mech_step(conn->context,
 					 c_conn->cparams,
@@ -563,8 +583,7 @@ int sasl_client_step(sasl_conn_t *conn,
        * 1. the mech does server-send-last
        * 2. the protocol does not
        * we need to return no data */
-      if(!(conn->flags & SASL_SUCCESS_DATA)
-	 && (c_conn->mech->plug->features & SASL_FEAT_WANT_SERVER_LAST)) {
+      if(!*clientout && !(conn->flags & SASL_SUCCESS_DATA)) {
 	  *clientout = "";
 	  *clientoutlen = 0;
       }
@@ -572,10 +591,172 @@ int sasl_client_step(sasl_conn_t *conn,
       if(!conn->oparams.maxoutbuf) {
 	  conn->oparams.maxoutbuf = conn->props.maxbufsize;
       }
-  }
-  
+
+      if(conn->oparams.user == NULL || conn->oparams.authid == NULL) {
+	  sasl_seterror(conn, 0,
+			"mech did not call canon_user for both authzid and authid");
+	  result = SASL_BADPROT;
+      }
+  }  
 
   RETURN(conn,result);
 }
 
+/* returns the length of all the mechanisms
+ * added up 
+ */
 
+static unsigned mech_names_len()
+{
+  cmechanism_t *listptr;
+  unsigned result = 0;
+
+  for (listptr = cmechlist->mech_list;
+       listptr;
+       listptr = listptr->next)
+    result += strlen(listptr->plug->mech_name);
+
+  return result;
+}
+
+
+int _sasl_client_listmech(sasl_conn_t *conn,
+			  const char *prefix,
+			  const char *sep,
+			  const char *suffix,
+			  const char **result,
+			  size_t *plen,
+			  int *pcount)
+{
+    cmechanism_t *m=NULL;
+    sasl_ssf_t minssf = 0;
+    int ret;
+    unsigned int resultlen;
+    int flag;
+    const char *mysep;
+
+    if(_sasl_client_active == 0) return SASL_NOTINIT;
+    if (!conn) return SASL_BADPARAM;
+    if(conn->type != SASL_CONN_CLIENT) PARAMERROR(conn);
+    
+    if (! result)
+	PARAMERROR(conn);
+    
+    if (plen != NULL)
+	*plen = 0;
+    if (pcount != NULL)
+	*pcount = 0;
+
+    if (sep) {
+	mysep = sep;
+    } else {
+	mysep = " ";
+    }
+
+    if(conn->props.min_ssf < conn->external.ssf) {
+	minssf = 0;
+    } else {
+	minssf = conn->props.min_ssf - conn->external.ssf;
+    }
+
+    if (! cmechlist || cmechlist->mech_length <= 0)
+	INTERROR(conn, SASL_NOMECH);
+
+    resultlen = (prefix ? strlen(prefix) : 0)
+	+ (strlen(mysep) * (cmechlist->mech_length - 1))
+	+ mech_names_len()
+	+ (suffix ? strlen(suffix) : 0)
+	+ 1;
+    ret = _buf_alloc(&conn->mechlist_buf,
+		     &conn->mechlist_buf_len, resultlen);
+    if(ret != SASL_OK) MEMERROR(conn);
+
+    if (prefix)
+	strcpy (conn->mechlist_buf,prefix);
+    else
+	*(conn->mechlist_buf) = '\0';
+
+    flag = 0;
+    for (m = cmechlist->mech_list; m != NULL; m = m->next) {
+	    /* do we have the prompts for it? */
+	    if (!have_prompts(conn, m->plug))
+		continue;
+
+	    /* is it strong enough? */
+	    if (minssf > m->plug->max_ssf)
+		continue;
+
+	    /* does it meet our security properties? */
+	    if (((conn->props.security_flags ^ m->plug->security_flags)
+		 & conn->props.security_flags) != 0) {
+		continue;
+	    }
+
+	    /* Can we meet it's features? */
+	    if ((m->plug->features & SASL_FEAT_NEEDSERVERFQDN)
+		&& !conn->serverFQDN) {
+		continue;
+	    }
+
+	    /* Okay, we like it, add it to the list! */
+
+	    if (pcount != NULL)
+		(*pcount)++;
+
+	    /* print seperator */
+	    if (flag) {
+		strcat(conn->mechlist_buf, mysep);
+	    } else {
+		flag = 1;
+	    }
+	    
+	    /* now print the mechanism name */
+	    strcat(conn->mechlist_buf, m->plug->mech_name);
+    }
+    
+  if (suffix)
+      strcat(conn->mechlist_buf,suffix);
+
+  if (plen!=NULL)
+      *plen=strlen(conn->mechlist_buf);
+
+  *result = conn->mechlist_buf;
+
+  return SASL_OK;
+}
+
+sasl_string_list_t *_sasl_client_mechs(void) 
+{
+  cmechanism_t *listptr;
+  sasl_string_list_t *retval = NULL, *next=NULL;
+
+  if(!_sasl_client_active) return NULL;
+
+  /* make list */
+  for (listptr = cmechlist->mech_list; listptr; listptr = listptr->next) {
+      next = sasl_ALLOC(sizeof(sasl_string_list_t));
+
+      if(!next && !retval) return NULL;
+      else if(!next) {
+	  next = retval->next;
+	  do {
+	      sasl_FREE(retval);
+	      retval = next;
+	      next = retval->next;
+	  } while(next);
+	  return NULL;
+      }
+      
+      next->d = listptr->plug->mech_name;
+
+      if(!retval) {
+	  next->next = NULL;
+	  retval = next;
+      } else {
+	  next->next = retval;
+	  retval = next;
+      }
+  }
+
+  return retval;
+}
