@@ -1,7 +1,7 @@
 /* SASL server API implementation
  * Rob Siemborski
  * Tim Martin
- * $Id: server.c,v 1.89 2001/12/04 02:05:27 rjs3 Exp $
+ * $Id: server.c,v 1.110 2002/05/05 14:13:48 ken3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -238,9 +238,6 @@ static void server_dispose(sasl_conn_t *pconn)
   if (s_conn->sparams->propctx)
       prop_dispose(&s_conn->sparams->propctx);
 
-  if (s_conn->mechlist_buf)
-      sasl_FREE(s_conn->mechlist_buf);
-
   if (s_conn->user_realm)
       sasl_FREE(s_conn->user_realm);
 
@@ -349,10 +346,6 @@ static void server_done(void) {
 	      prevm->plug->mech_free(prevm->plug->glob_context,
 				     mechlist->utils);
 	  }
-	  
-	  if (prevm->plug->glob_context!=NULL) {
-	      sasl_FREE(prevm->plug->glob_context);
-	  }
 
 	  sasl_FREE(prevm->plugname);	  	  
 	  sasl_FREE(prevm);    
@@ -498,10 +491,10 @@ struct secflag_map_s secflag_map[] = {
     { "forward_secrecy", SASL_SEC_FORWARD_SECRECY },
     { "noanonymous", SASL_SEC_NOANONYMOUS },
     { "pass_credentials", SASL_SEC_PASS_CREDENTIALS },
+    { "mutual_auth", SASL_SEC_MUTUAL_AUTH },
     { NULL, 0x0 }
 };
 
-
 static int parse_mechlist_file(const char *mechlistfile)
 {
     FILE *f;
@@ -555,7 +548,6 @@ static int parse_mechlist_file(const char *mechlistfile)
 		_sasl_log(NULL, SASL_LOG_ERR,
 			  "%s: couldn't identify flag '%s'",
 			  nplug->mech_name, t);
-
 	    }
 	    free(t);
 	}
@@ -632,20 +624,22 @@ int sasl_server_init(const sasl_callback_t *callbacks,
     }
 
     /* load internal plugins */
-    sasl_server_add_plugin("EXTERNAL", &external_server_init);
+    sasl_server_add_plugin("EXTERNAL", &external_server_plug_init);
 
     /* delayed loading of plugins? (DSO only, as it doesn't
      * make much [any] sense to delay in the static library case) */
     if (!_is_sasl_server_static &&
 	_sasl_getcallback(NULL, SASL_CB_GETOPT, &getopt, &context) 
 	   == SASL_OK) {
-	getopt(context, NULL, "plugin_list", &pluginfile, NULL);
+	/* No sasl_conn_t was given to getcallback, so we provide the
+	 * global callbacks structure */
+	ret = getopt(&global_callbacks, NULL, "plugin_list", &pluginfile, NULL);
     }
     
     if (pluginfile != NULL) {
 	/* this file should contain a list of plugins available.
 	   we'll load on demand. */
-	
+
 	/* Ask the application if it's safe to use this file */
 	ret = ((sasl_verifyfile_t *)(vf->proc))(vf->context,
 						pluginfile,
@@ -664,12 +658,14 @@ int sasl_server_init(const sasl_callback_t *callbacks,
 				 _sasl_find_getpath_callback(callbacks),
 				 _sasl_find_verifyfile_callback(callbacks));
     }
-  
+
     if (ret == SASL_OK) {
 	/* _sasl_server_active shows if we're active or not. 
 	   server_done() sets it back to 0 */
 	_sasl_server_active = 1;
 	_sasl_server_idle_hook = &server_idle;
+
+	ret = _sasl_build_mechlist();
     }
 
     return ret;
@@ -753,6 +749,7 @@ int sasl_server_new(const char *service,
   sasl_server_conn_t *serverconn;
   sasl_utils_t *utils;
 
+  if (_sasl_server_active==0) return SASL_NOTINIT;
   if (! pconn) return SASL_FAIL;
   if (! service) return SASL_FAIL;
 
@@ -800,6 +797,7 @@ int sasl_server_new(const char *service,
   serverconn->sparams->canon_user = &_sasl_canon_user;
   serverconn->sparams->serverFQDN = (*pconn)->serverFQDN;
   serverconn->sparams->props = serverconn->base.props;
+  serverconn->sparams->flags = flags;
 
   /* set some variables */
   if (user_realm) {
@@ -834,7 +832,7 @@ static int mech_permitted(sasl_conn_t *conn,
     void *context;
     sasl_ssf_t minssf = 0;
 
-    if(!conn) return SASL_BADPARAM;
+    if(!conn) return 0;
 
     if(! mech || ! mech->plug) {
 	PARAMERROR(conn);
@@ -880,12 +878,12 @@ static int mech_permitted(sasl_conn_t *conn,
     for(cur = s_conn->mech_contexts; cur; cur=cur->next) {
 	if(cur->mech == mech) {
 	    /* If it's not mech_avail'd, then stop now */
-	    if(!cur->context) return SASL_NOMECH;
+	    if(!cur->context) return 0;
 	    break;
 	}
     }
     
-    if (plug == &external_server_mech) {
+    if (!strcasecmp(plug->mech_name, "EXTERNAL")) {
 	/* Special case for the external mechanism */
 	if (conn->props.min_ssf > conn->external.ssf
 	    || ! conn->external.auth_id) {
@@ -1036,10 +1034,12 @@ int sasl_server_start(sasl_conn_t *conn,
     sasl_server_conn_t *s_conn=(sasl_server_conn_t *) conn;
     int result;
     context_list_t *cur, **prev;
+    mechanism_t *m;
+
+    if (_sasl_server_active==0) return SASL_NOTINIT;
 
     /* make sure mech is valid mechanism
        if not return appropriate error */
-    mechanism_t *m;
     m=mechlist->mech_list;
 
     /* check parameters */
@@ -1048,6 +1048,9 @@ int sasl_server_start(sasl_conn_t *conn,
     if (!mech || ((clientin==NULL) && (clientinlen>0)))
 	PARAMERROR(conn);
 
+    if(serverout) *serverout = NULL;
+    if(serveroutlen) *serveroutlen = 0;
+
     while (m!=NULL)
     {
 	if ( strcasecmp(mech,m->plug->mech_name)==0)
@@ -1157,37 +1160,32 @@ int sasl_server_start(sasl_conn_t *conn,
     }
     
     if (result == SASL_OK) {
-	if(clientin &&
-	   (s_conn->mech->plug->features &
-	    (SASL_FEAT_WANT_CLIENT_FIRST | SASL_FEAT_INTERNAL_CLIENT_FIRST))) {
-	    /* both protocol & mech did client-send-first */
-	    result = sasl_server_step(conn,
-				      clientin, clientinlen,
-				      serverout, serveroutlen);
-	} else if (clientin) {
-	    /* Remote sent first, but mechanism does not support it.
-	     * RFC 2222 says we fail at this point. */
-	    sasl_seterror(conn, 0,
-			  "Remote sent first but mech does not allow it.");
-	    result = SASL_BADPROT;
-	} else {
-	    /* protocol CAN do client-send-first, but no data arrived */
-	    /* OR protocol can't do client-send-first */
-	    if(!(s_conn->mech->plug->features & SASL_FEAT_WANT_CLIENT_FIRST) ||
-	    (s_conn->mech->plug->features & SASL_FEAT_INTERNAL_CLIENT_FIRST)) {
-		/* Mech wants server-first, so let them have it */
-		/* or the send-first situation is handled internally
-		 *  by the mechanism */
-		result = sasl_server_step(conn,
-					  clientin, clientinlen,
-					  serverout, serveroutlen);
-	    } else {
-		/* Mech wants client first anyway, so we should do that */
-		*serverout = "";
-		*serveroutlen = 0;
-		result = SASL_CONTINUE;
-	    }
-	}	 
+         if(clientin) {
+            if(s_conn->mech->plug->features & SASL_FEAT_SERVER_FIRST) {
+                /* Remote sent first, but mechanism does not support it.
+                 * RFC 2222 says we fail at this point. */
+                sasl_seterror(conn, 0,
+                              "Remote sent first but mech does not allow it.");
+                result = SASL_BADPROT;
+            } else {
+                /* Mech wants client-first, so let them have it */
+                result = sasl_server_step(conn,
+                                          clientin, clientinlen,
+                                          serverout, serveroutlen);
+            }
+        } else {
+            if(s_conn->mech->plug->features & SASL_FEAT_WANT_CLIENT_FIRST) {
+                /* Mech wants client first anyway, so we should do that */
+                *serverout = "";
+                *serveroutlen = 0;
+                result = SASL_CONTINUE;
+            } else {
+                /* Mech wants server-first, so let them have it */
+                result = sasl_server_step(conn,
+                                          clientin, clientinlen,
+                                          serverout, serveroutlen);
+            }
+	}
     }
 
  done:
@@ -1232,6 +1230,7 @@ int sasl_server_step(sasl_conn_t *conn,
     sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;  /* cast */
 
     /* check parameters */
+    if (_sasl_server_active==0) return SASL_NOTINIT;
     if (!conn) return SASL_BADPARAM;
     if ((clientin==NULL) && (clientinlen>0))
 	PARAMERROR(conn);
@@ -1241,6 +1240,15 @@ int sasl_server_step(sasl_conn_t *conn,
 	return SASL_OK;
     }
 
+    /* Don't do another step if the plugin told us that we're done */
+    if (conn->oparams.doneflag) {
+	_sasl_log(conn, SASL_LOG_ERR, "attempting server step after doneflag");
+	return SASL_FAIL;
+    }
+
+    if(serverout) *serverout = NULL;
+    if(serveroutlen) *serveroutlen = 0;
+
     ret = s_conn->mech->plug->mech_step(conn->context,
 					s_conn->sparams,
 					clientin,
@@ -1260,8 +1268,7 @@ int sasl_server_step(sasl_conn_t *conn,
 	 *
 	 * in this case, return SASL_CONTINUE and remember we are done.
 	 */
-	if(!(conn->flags & SASL_SUCCESS_DATA)
-	   && (s_conn->mech->plug->features & SASL_FEAT_WANT_SERVER_LAST)) {
+	if(*serverout && !(conn->flags & SASL_SUCCESS_DATA)) {
 	    s_conn->sent_last = 1;
 	    ret = SASL_CONTINUE;
 	}
@@ -1269,12 +1276,11 @@ int sasl_server_step(sasl_conn_t *conn,
 	    conn->oparams.maxoutbuf = conn->props.maxbufsize;
 	}
 
-	/* If authorization id is not the same as authenticationid, get
-	 * the real auxprops */
-	if(strcmp(conn->oparams.user, conn->oparams.authid)) {
-	    prop_clear(s_conn->sparams->propctx,0);
-	    _sasl_auxprop_lookup(s_conn->sparams, 0,
-				 conn->oparams.user, conn->oparams.ulen);
+	if(conn->oparams.user == NULL || conn->oparams.authid == NULL) {
+	    sasl_seterror(conn, 0,
+			  "mech did not call canon_user for both authzid " \
+			  "and authid");
+	    ret = SASL_BADPROT;
 	}	
     }
     
@@ -1309,36 +1315,18 @@ static unsigned mech_names_len()
 }
 
 /* This returns a list of mechanisms in a NUL-terminated string
- *  user          -- restricts mechanisms to those available to that user
- *                   (may be NULL)
- *  prefix        -- appended to beginning of result
- *  sep           -- appended between mechanisms
- *  suffix        -- appended to end of result
- * results:
- *  result        -- NUL terminated allocated result, only good until
- *                   next call to sasl_listmech
- *  plen          -- gets length of result (excluding NUL), may be NULL
- *  pcount        -- gets number of mechanisms, may be NULL
  *
- * returns:
- *  SASL_OK        -- success
- *  SASL_NOMEM     -- not enough memory
- *  SASL_NOMECH    -- no enabled mechanisms
- */
-
-/*
  * The default behavior is to seperate with spaces if sep==NULL
  */
-int sasl_listmech(sasl_conn_t *conn,
-		  const char *user __attribute__((unused)),
-		  const char *prefix,
-		  const char *sep,
-		  const char *suffix,
-		  const char **result,
-		  unsigned *plen,
-		  int *pcount)
+int _sasl_server_listmech(sasl_conn_t *conn,
+			  const char *user __attribute__((unused)),
+			  const char *prefix,
+			  const char *sep,
+			  const char *suffix,
+			  const char **result,
+			  size_t *plen,
+			  int *pcount)
 {
-  sasl_server_conn_t *s_conn = (sasl_server_conn_t *)conn;
   int lup;
   mechanism_t *listptr;
   int ret;
@@ -1349,6 +1337,7 @@ int sasl_listmech(sasl_conn_t *conn,
   /* if there hasn't been a sasl_sever_init() fail */
   if (_sasl_server_active==0) return SASL_NOTINIT;
   if (!conn) return SASL_BADPARAM;
+  if (conn->type != SASL_CONN_SERVER) PARAMERROR(conn);
   
   if (! result)
       PARAMERROR(conn);
@@ -1372,14 +1361,14 @@ int sasl_listmech(sasl_conn_t *conn,
 	    + mech_names_len()
             + (suffix ? strlen(suffix) : 0)
 	    + 1;
-  ret = _buf_alloc(&s_conn->mechlist_buf,
-		   &s_conn->mechlist_buf_len, resultlen);
+  ret = _buf_alloc(&conn->mechlist_buf,
+		   &conn->mechlist_buf_len, resultlen);
   if(ret != SASL_OK) MEMERROR(conn);
 
   if (prefix)
-    strcpy (s_conn->mechlist_buf,prefix);
+    strcpy (conn->mechlist_buf,prefix);
   else
-    *(s_conn->mechlist_buf) = '\0';
+    *(conn->mechlist_buf) = '\0';
 
   listptr = mechlist->mech_list;  
    
@@ -1393,30 +1382,66 @@ int sasl_listmech(sasl_conn_t *conn,
 
 	  /* print seperator */
 	  if (flag) {
-	      strcat(s_conn->mechlist_buf, mysep);
+	      strcat(conn->mechlist_buf, mysep);
 	  } else {
 	      flag = 1;
 	  }
 
 	  /* now print the mechanism name */
-	  strcat(s_conn->mechlist_buf, listptr->plug->mech_name);
+	  strcat(conn->mechlist_buf, listptr->plug->mech_name);
       }
 
       listptr = listptr->next;
   }
 
   if (suffix)
-      strcat(s_conn->mechlist_buf,suffix);
+      strcat(conn->mechlist_buf,suffix);
 
   if (plen!=NULL)
-      *plen=strlen(s_conn->mechlist_buf);
+      *plen=strlen(conn->mechlist_buf);
 
-  *result = s_conn->mechlist_buf;
+  *result = conn->mechlist_buf;
 
   return SASL_OK;
   
 }
 
+sasl_string_list_t *_sasl_server_mechs(void) 
+{
+  mechanism_t *listptr;
+  sasl_string_list_t *retval = NULL, *next=NULL;
+
+  if(!_sasl_server_active) return NULL;
+
+  /* make list */
+  for (listptr = mechlist->mech_list; listptr; listptr = listptr->next) {
+      next = sasl_ALLOC(sizeof(sasl_string_list_t));
+
+      if(!next && !retval) return NULL;
+      else if(!next) {
+	  next = retval->next;
+	  do {
+	      sasl_FREE(retval);
+	      retval = next;
+	      next = retval->next;
+	  } while(next);
+	  return NULL;
+      }
+      
+      next->d = listptr->plug->mech_name;
+
+      if(!retval) {
+	  next->next = NULL;
+	  retval = next;
+      } else {
+	  next->next = retval;
+	  retval = next;
+      }
+  }
+
+  return retval;
+}
+
 #define EOSTR(s,n) (((s)[n] == '\0') || ((s)[n] == ' ') || ((s)[n] == '\t'))
 static int is_mech(const char *t, const char *m)
 {
@@ -1433,7 +1458,7 @@ static int _sasl_checkpass(sasl_conn_t *conn, const char *service,
     sasl_getopt_t *getopt;
     sasl_server_userdb_checkpass_t *checkpass_cb;
     void *context;
-    const char *mech;
+    const char *mlist, *mech;
     struct sasl_verify_password_s *v;
 
     /* call userdb callback function, if available */
@@ -1449,23 +1474,32 @@ static int _sasl_checkpass(sasl_conn_t *conn, const char *service,
     /* figure out how to check (i.e. auxprop or saslauthd or pwcheck) */
     if (_sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context)
             == SASL_OK) {
-        getopt(context, NULL, "pwcheck_method", &mech, NULL);
+        getopt(context, NULL, "pwcheck_method", &mlist, NULL);
     }
 
-    if(!mech) mech = DEFAULT_CHECKPASS_MECH;
+    if(!mlist) mlist = DEFAULT_CHECKPASS_MECH;
 
     result = SASL_NOMECH;
 
-    for (v = _sasl_verify_password; v->name; v++) {
-	if(is_mech(mech, v->name)) {
-	    result = v->verify(conn, user, pass, service, s_conn->user_realm);
-	    break;
+    mech = mlist;
+    while (*mech && result != SASL_OK) {
+	for (v = _sasl_verify_password; v->name; v++) {
+	    if(is_mech(mech, v->name)) {
+		result = v->verify(conn, user, pass, service,
+				   s_conn->user_realm);
+		break;
+	    }
+	}
+	if (result != SASL_OK) {
+	    /* skip to next mech in list */
+	    while (*mech && !isspace((int) *mech)) mech++;
+	    while (*mech && isspace((int) *mech)) mech++;
 	}
     }
 
     if (result == SASL_NOMECH) {
 	/* no mechanism available ?!? */
-	_sasl_log(conn, SASL_LOG_ERR, "unkwnown password verifier %s", mech);
+	_sasl_log(conn, SASL_LOG_ERR, "unknown password verifier %s", mech);
     }
 
     if (result != SASL_OK)
@@ -1538,7 +1572,7 @@ int sasl_user_exists(sasl_conn_t *conn,
 		     const char *user) 
 {
     int result=SASL_NOMECH;
-    const char *mech;
+    const char *mlist, *mech;
     void *context;
     sasl_getopt_t *getopt;
     struct sasl_verify_password_s *v;
@@ -1554,17 +1588,25 @@ int sasl_user_exists(sasl_conn_t *conn,
     /* figure out how to check (i.e. auxprop or saslauthd or pwcheck) */
     if (_sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context)
             == SASL_OK) {
-        getopt(context, NULL, "pwcheck_method", &mech, NULL);
+        getopt(context, NULL, "pwcheck_method", &mlist, NULL);
     }
 
-    if(!mech) mech = DEFAULT_CHECKPASS_MECH;
+    if(!mlist) mlist = DEFAULT_CHECKPASS_MECH;
 
     result = SASL_NOMECH;
 
-    for (v = _sasl_verify_password; v->name; v++) {
-	if(is_mech(mech, v->name)) {
-	    result = v->verify(conn, user, NULL, service, user_realm);
-	    break;
+    mech = mlist;
+    while (*mech && result != SASL_OK) {
+	for (v = _sasl_verify_password; v->name; v++) {
+	    if(is_mech(mech, v->name)) {
+		result = v->verify(conn, user, NULL, service, user_realm);
+		break;
+	    }
+	}
+	if (result != SASL_OK) {
+	    /* skip to next mech in list */
+	    while (*mech && !isspace((int) *mech)) mech++;
+	    while (*mech && isspace((int) *mech)) mech++;
 	}
     }
 
@@ -1655,17 +1697,22 @@ int sasl_checkapop(sasl_conn_t *conn,
         RETURN(conn, result);
 
     /* Cannonify it */
-    result = _sasl_canon_user(conn,
-			   user, user_len,
-			   user, user_len,
-			   0, &(conn->oparams));
+    result = _sasl_canon_user(conn, user, user_len,
+	                      SASL_CU_AUTHZID, &(conn->oparams));
+    if(result != SASL_OK) 
+    {
+        sasl_FREE(user);
+        RETURN(conn, result);
+    }
 
+    result = _sasl_canon_user(conn, user, user_len, SASL_CU_AUTHID,
+	                      &(conn->oparams));
     sasl_FREE(user);
 
     if(result != SASL_OK) RETURN(conn, result);
 
     /* Do APOP verification */
-    result = _sasl_auxprop_verify_apop(conn, conn->oparams.user,
+    result = _sasl_auxprop_verify_apop(conn, conn->oparams.authid,
 	challenge, user_end + 1, s_conn->user_realm);
 
     /* If verification failed, we don't want to encourage getprop to work */
