@@ -1,7 +1,7 @@
 /* testsuite.c -- Stress the library a little
  * Rob Siemborski
  * Tim Martin
- * $Id: testsuite.c,v 1.15 2001/12/04 02:07:02 rjs3 Exp $
+ * $Id: testsuite.c,v 1.25 2002/05/06 21:00:37 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -132,12 +132,94 @@ const char *corrupt_types[] = {
     "CORRUPT_SIZE"
 };
 
+void fatal(char *str)
+{
+    printf("Failed with: %s\n",str);
+    exit(3);
+}
+
+/* interactions we support */
+static sasl_callback_t client_interactions[] = {
+  {
+    SASL_CB_GETREALM, NULL, NULL
+  }, {
+    SASL_CB_USER, NULL, NULL
+  }, {
+    SASL_CB_AUTHNAME, NULL, NULL
+  }, {
+    SASL_CB_PASS, NULL, NULL    
+  }, {
+    SASL_CB_LIST_END, NULL, NULL
+  }
+};
+
+int test_getrealm(void *context __attribute__((unused)), int id,
+		  const char **availrealms __attribute__((unused)),
+		  const char **result) 
+{
+    if(id != SASL_CB_GETREALM) fatal("test_getrealm not looking for realm");
+    if(!result) return SASL_BADPARAM;
+    *result = myhostname;
+    return SASL_OK;
+}
+
+/* yeah, yeah, this leaks memory, but we don't really care in
+ * the testsuite */
+int test_getsecret(sasl_conn_t *conn __attribute__((unused)),
+		   void *context __attribute__((unused)), int id,
+		   sasl_secret_t **psecret) 
+{
+    if(id != SASL_CB_PASS) fatal("test_getsecret not looking for pass");
+    if(!psecret) return SASL_BADPARAM;
+
+    *psecret = malloc(sizeof(sasl_secret_t) + strlen(password));
+    (*psecret)->len = strlen(password);
+    strcpy((*psecret)->data, password);
+
+    return SASL_OK;
+}
+
+int test_getsimple(void *context __attribute__((unused)), int id,
+		   const char **result, unsigned *len) 
+{
+    if(!result) return SASL_BADPARAM;
+    
+    if (id==SASL_CB_USER && proxyflag == 0) {
+	*result=(char *) username;
+    } else if (id==SASL_CB_USER && proxyflag == 1) {
+	*result=(char *) proxyasname;
+    } else if (id==SASL_CB_AUTHNAME) {
+	*result=(char *) authname;
+    } else {
+	printf("I want %d\n", id);
+	fatal("unknown callback in test_getsimple");
+    }
+
+    if(len) *len = strlen(*result);
+    return SASL_OK;
+}
+
+/* callbacks we support */
+static sasl_callback_t client_callbacks[] = {
+  {
+    SASL_CB_GETREALM, test_getrealm, NULL
+  }, {
+    SASL_CB_USER, test_getsimple, NULL
+  }, {
+    SASL_CB_AUTHNAME, test_getsimple, NULL
+  }, {
+    SASL_CB_PASS, test_getsecret, NULL    
+  }, {
+    SASL_CB_LIST_END, NULL, NULL
+  }
+};
 
 typedef void *foreach_t(char *mech, void *rock);
 
 typedef struct tosend_s {
     corrupt_type_t type; /* type of corruption to make */
     int step; /* step it should send bogus data on */
+    sasl_callback_t *client_callbacks; /* which client callbacks to use */
 } tosend_t;
 
 typedef struct mem_info 
@@ -293,12 +375,6 @@ int mem_stat()
 
 /************* End Memory Allocation functions ******/
 
-void fatal(char *str)
-{
-    printf("Failed with: %s\n",str);
-    exit(3);
-}
-
 /* my mutex functions */
 int g_mutex_cnt = 0;
 
@@ -513,11 +589,86 @@ void test_init(void)
     if (result!=SASL_OK) fatal("Didn't deal with ok callback path very well");
     sasl_done();
 
+    /* and the client */
+    result = sasl_client_init(withokpathsasl_cb);
+
+    if (result!=SASL_OK)
+	fatal("Client didn't deal with ok callback path very well");
+    sasl_done();
+
     /* try giving it an invalid path for where the plugins are */
     result = sasl_server_init(withbadpathsasl_cb, NULL);
 
     if (result==SASL_OK) fatal("Allowed invalid path");
     sasl_done();
+
+    /* and the client - xxx is this necessary?*/
+#if 0
+    result = sasl_client_init(withbadpathsasl_cb);
+
+    if (result==SASL_OK)
+	fatal("Client allowed invalid path");
+    sasl_done();
+#endif
+
+    /* Now try to break all the sasl_server_* functions for not returning
+       SASL_NOTINIT */
+
+    if(sasl_global_listmech())
+	fatal("sasl_global_listmech did not return NULL with no library initialized");
+
+    if(sasl_server_new(NULL, NULL, NULL, NULL, NULL, NULL, 0, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_server_new did not return SASL_NOTINIT");
+
+/* Can't check this validly without a server conn, so this would be
+   a hard one to tickle anyway */
+#if 0    
+    if(sasl_listmech(NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_listmech did not return SASL_NOTINIT");
+#endif
+
+    if(sasl_server_start(NULL, NULL, NULL, 0, NULL, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_server_start did not return SASL_NOTINIT");
+
+    if(sasl_server_step(NULL, NULL, 0, NULL, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_server_step did not return SASL_NOTINIT");
+    
+#ifdef DO_SASL_CHECKAPOP
+    if(sasl_checkapop(NULL, NULL, 0, NULL, 0)
+       != SASL_NOTINIT)
+	fatal("sasl_checkapop did not return SASL_NOTINIT");
+#endif    
+
+    if(sasl_checkpass(NULL, NULL, 0, NULL, 0)
+       != SASL_NOTINIT)
+	fatal("sasl_checkpass did not return SASL_NOTINIT");
+    
+    if(sasl_user_exists(NULL, NULL, NULL, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_user_exists did not return SASL_NOTINIT");
+
+    if(sasl_setpass(NULL, NULL, NULL, 0, NULL, 0, 0)
+       != SASL_NOTINIT)
+	fatal("sasl_setpass did not return SASL_NOTINIT");
+
+    /* And sasl_client_*... */
+
+    if(sasl_client_new(NULL, NULL, NULL, NULL, NULL, 0, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_client_new did not return SASL_NOTINIT");
+
+    if(sasl_client_start(NULL, NULL, NULL, NULL, NULL, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_client_start did not return SASL_NOTINIT");
+
+    if(sasl_client_step(NULL, NULL, 0, NULL, NULL, NULL)
+       != SASL_NOTINIT)
+	fatal("sasl_client_step did not return SASL_NOTINIT");
+
 }
 
 
@@ -527,11 +678,12 @@ void test_init(void)
 
 void test_listmech(void)
 {
-    sasl_conn_t *saslconn;
+    sasl_conn_t *saslconn, *cconn;
     int result;
     const char *str = NULL;
     unsigned int plen;
-    unsigned lup, pcount;
+    unsigned lup, flag, pcount;
+    const char **list;
 
     /* test without initializing library */
     result = sasl_listmech(NULL, /* conn */
@@ -546,18 +698,57 @@ void test_listmech(void)
     /*    printf("List mech without library initialized: %s\n",sasl_errstring(result,NULL,NULL));*/
     if (result == SASL_OK) fatal("Failed sasl_listmech() with NULL saslconn");
 
-
-
-
     if (sasl_server_init(emptysasl_cb,"TestSuite")!=SASL_OK)
 	fatal("can't sasl_server_init");
+    if (sasl_client_init(client_interactions)!=SASL_OK)
+	fatal("can't sasl_client_init");
 
     if (sasl_server_new("rcmd", myhostname,
 			NULL, NULL, NULL, NULL, 0, 
-			&saslconn) != SASL_OK) {
+			&saslconn) != SASL_OK)
 	fatal("can't sasl_server_new");
-    }
 
+    if (sasl_setprop(saslconn, SASL_AUTH_EXTERNAL, authname)!=SASL_OK)
+	fatal("sasl_setprop(SASL_AUTH_EXTERNAL) failed");
+
+    /* client new connection */
+    if (sasl_client_new("rcmd",
+			myhostname,
+			NULL, NULL, NULL,
+			0,
+			&cconn)!= SASL_OK)
+	fatal("sasl_client_new() failure");
+
+    if (sasl_setprop(cconn, SASL_AUTH_EXTERNAL, authname)!=SASL_OK)
+	fatal("sasl_setprop(SASL_AUTH_EXTERNAL) failed");
+
+    /* try both sides */
+    list = sasl_global_listmech();
+    if(!list) fatal("sasl_global_listmech failure");
+
+    printf(" [");
+    flag = 0;
+    for(lup = 0; list[lup]; lup++) {
+	if(flag) printf(",");
+	else flag++;
+	printf("%s",list[lup]);
+    }
+    printf("]\n");
+
+    /* try client side */
+    result = sasl_listmech(cconn,
+			   NULL,
+			   " [",
+			   ",",
+			   "]",
+			   &str,
+			   NULL,
+			   NULL);
+    if(result == SASL_OK) {
+	printf("Client mechlist:\n%s\n", str);
+    } else {
+	fatal("client side sasl_listmech failed");
+    }
 
     /* Test with really long user */
 
@@ -656,6 +847,7 @@ void test_listmech(void)
 
     /* Call sasl done then make sure listmech doesn't work anymore */
     sasl_dispose(&saslconn);
+    sasl_dispose(&cconn);
     sasl_done();
 
     result = sasl_listmech(saslconn,
@@ -855,23 +1047,6 @@ void test_props(void)
 	fatal("ctx not null after prop_dispose");
 }
 
-/* callbacks we support */
-static sasl_callback_t client_callbacks[] = {
-  {
-#ifdef SASL_CB_GETREALM
-    SASL_CB_GETREALM, NULL, NULL
-  }, {
-#endif
-    SASL_CB_USER, NULL, NULL
-  }, {
-    SASL_CB_AUTHNAME, NULL, NULL
-  }, {
-    SASL_CB_PASS, NULL, NULL    
-  }, {
-    SASL_CB_LIST_END, NULL, NULL
-  }
-};
-
 void interaction (int id, const char *prompt,
 		  const char **tresult, unsigned int *tlen)
 {
@@ -883,10 +1058,8 @@ void interaction (int id, const char *prompt,
 	*tresult=(char *) proxyasname;
     } else if (id==SASL_CB_AUTHNAME) {
 	*tresult=(char *) authname;
-#ifdef SASL_CB_GETREALM
     } else if ((id==SASL_CB_GETREALM)) {
 	*tresult=(char *) myhostname;
-#endif
     } else {
 	int c;
 	
@@ -930,6 +1103,9 @@ void set_properties(sasl_conn_t *conn, const sasl_security_properties_t *props)
        	if (sasl_setprop(conn, SASL_SEC_PROPS, props)!=SASL_OK)
 	    fatal("sasl_setprop() failed");
     }
+
+    if (sasl_setprop(conn, SASL_AUTH_EXTERNAL, authname)!=SASL_OK)
+	fatal("sasl_setprop(SASL_AUTH_EXTERNAL) failed");
 }
 
 /*
@@ -1050,12 +1226,13 @@ void sendbadsecond(char *mech, void *rock)
     struct sockaddr_in addr;
     struct hostent *hp;
     char buf[8192];
+    int reauth = 1;
 
     printf("%s --> start\n",mech);
     
     if (strcmp(mech,"GSSAPI")==0) service = gssapi_service;
 
-    if (sasl_client_init(client_callbacks)!=SASL_OK) fatal("Unable to init client");
+    if (sasl_client_init(client_interactions)!=SASL_OK) fatal("Unable to init client");
 
     if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK) fatal("unable to init server");
 
@@ -1068,6 +1245,7 @@ void sendbadsecond(char *mech, void *rock)
     memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);
     addr.sin_port = htons(0);
 
+ reauth: /* loop back for reauth testing */
     sprintf(buf,"%s;%d", inet_ntoa(addr.sin_addr), 23);
 
     /* client new connection */
@@ -1217,6 +1395,13 @@ void sendbadsecond(char *mech, void *rock)
 	    fatal("client was not ok on last server step");
     }
     
+    if (reauth) {
+ 	sasl_dispose(&clientconn);
+ 	sasl_dispose(&saslconn);
+ 
+ 	reauth = 0;
+ 	goto reauth;
+    }
 
     /* client to server */
     result = sasl_encode(clientconn, CLIENT_TO_SERVER,
@@ -1278,7 +1463,8 @@ void sendbadsecond(char *mech, void *rock)
 /* Authenticate two sasl_conn_t's to eachother, validly.
  * used to test the security layer */
 int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn,
-           const sasl_security_properties_t *props, int fail_ok)
+           const sasl_security_properties_t *props,
+	   sasl_callback_t *c_calls, int fail_ok)
 {
     int result, need_another_client = 0;
     sasl_conn_t *saslconn;
@@ -1296,7 +1482,7 @@ int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn,
     
     if (strcmp(mech,"GSSAPI")==0) service = gssapi_service;
 
-    result = sasl_client_init(client_callbacks);
+    result = sasl_client_init((c_calls ? c_calls : client_interactions));
     if (result!=SASL_OK) {
 	if(!fail_ok) fatal("Unable to init client");
 	else return result;
@@ -1431,7 +1617,8 @@ int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn,
  * without allowing client-send-first */
 int doauth_noclientfirst(char *mech, sasl_conn_t **server_conn,
 			 sasl_conn_t **client_conn,
-			 const sasl_security_properties_t *props)
+			 const sasl_security_properties_t *props,
+			 sasl_callback_t *c_calls)
 {
     int result, need_another_client = 0;
     sasl_conn_t *saslconn;
@@ -1451,7 +1638,7 @@ int doauth_noclientfirst(char *mech, sasl_conn_t **server_conn,
     if (strcmp(mech,"GSSAPI")==0) service = gssapi_service;
 
 
-    if (sasl_client_init(client_callbacks)!=SASL_OK)
+    if (sasl_client_init((c_calls ? c_calls : client_interactions))!=SASL_OK)
 	fatal("Unable to init client");
 
     if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK)
@@ -1563,7 +1750,8 @@ int doauth_noclientfirst(char *mech, sasl_conn_t **server_conn,
  * used to test the security layer */
 int doauth_serverlast(char *mech, sasl_conn_t **server_conn,
 		      sasl_conn_t **client_conn,
-		      const sasl_security_properties_t *props)
+		      const sasl_security_properties_t *props,
+		      sasl_callback_t *c_calls)
 {
     int result, need_another_client = 0;
     sasl_conn_t *saslconn;
@@ -1582,9 +1770,11 @@ int doauth_serverlast(char *mech, sasl_conn_t **server_conn,
     
     if (strcmp(mech,"GSSAPI")==0) service = gssapi_service;
 
-    if (sasl_client_init(client_callbacks)!=SASL_OK) fatal("Unable to init client");
+    if (sasl_client_init((c_calls ? c_calls : client_interactions))!=SASL_OK)
+	fatal("unable to init client");
 
-    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK) fatal("unable to init server");
+    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK)
+	fatal("unable to init server");
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
 	perror("gethostbyname");
@@ -1698,7 +1888,8 @@ int doauth_serverlast(char *mech, sasl_conn_t **server_conn,
  * without allowing client-send-first */
 int doauth_noclientfirst_andserverlast(char *mech, sasl_conn_t **server_conn,
 				       sasl_conn_t **client_conn,
-				       const sasl_security_properties_t *props)
+				       const sasl_security_properties_t *props,
+				       sasl_callback_t *c_calls)
 {
     int result, need_another_client = 0;
     sasl_conn_t *saslconn;
@@ -1717,9 +1908,11 @@ int doauth_noclientfirst_andserverlast(char *mech, sasl_conn_t **server_conn,
     
     if (strcmp(mech,"GSSAPI")==0) service = gssapi_service;
 
-    if (sasl_client_init(client_callbacks)!=SASL_OK) fatal("Unable to init client");
+    if (sasl_client_init((c_calls ? c_calls : client_interactions))!=SASL_OK)
+	fatal("unable to init client");
 
-    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK) fatal("unable to init server");
+    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK)
+	fatal("unable to init server");
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
 	perror("gethostbyname");
@@ -1854,6 +2047,15 @@ const sasl_security_properties_t force_des = {
     NULL	    
 };
 
+const sasl_security_properties_t force_rc4_56 = {
+    0,
+    56,
+    8192,
+    0,
+    NULL,
+    NULL	    
+};
+
 const sasl_security_properties_t force_3des = {
     0,
     112,
@@ -1889,7 +2091,7 @@ void do_proxypolicy_test(char *mech, void *rock __attribute__((unused)))
     
     printf("%s --> start\n", mech);
     proxyflag = 1;
-    if(doauth(mech, &sconn, &cconn, &security_props, 0) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, &security_props, NULL, 0) != SASL_OK) {
 	fatal("doauth failed in do_proxypolicy_test");
     }
 
@@ -1907,14 +2109,16 @@ void do_proxypolicy_test(char *mech, void *rock __attribute__((unused)))
     printf("%s --> successful result\n",mech);
 }
 
-void test_clientfirst(char *mech, void *rock __attribute__((unused))) 
+void test_clientfirst(char *mech, void *rock) 
 {
     sasl_conn_t *sconn, *cconn;
+    tosend_t *tosend = (tosend_t *)rock;
     
     printf("%s --> start\n", mech);
 
     /* Basic crash-tests (none should cause a crash): */
-    if(doauth(mech, &sconn, &cconn, &security_props, 0) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, &security_props, tosend->client_callbacks,
+	      0) != SASL_OK) {
 	fatal("doauth failed in test_clientfirst");
     }
 
@@ -1923,15 +2127,16 @@ void test_clientfirst(char *mech, void *rock __attribute__((unused)))
     printf("%s --> successful result\n", mech);
 }
 
-void test_noclientfirst(char *mech, void *rock __attribute__((unused))) 
+void test_noclientfirst(char *mech, void *rock) 
 {
     sasl_conn_t *sconn, *cconn;
+    tosend_t *tosend = (tosend_t *)rock;
     
     printf("%s --> start\n", mech);
 
     /* Basic crash-tests (none should cause a crash): */
-    if(doauth_noclientfirst(mech, &sconn, &cconn, &security_props)
-       != SASL_OK) {
+    if(doauth_noclientfirst(mech, &sconn, &cconn, &security_props,
+	tosend->client_callbacks) != SASL_OK) {
 	fatal("doauth failed in test_noclientfirst");
     }
 
@@ -1940,14 +2145,16 @@ void test_noclientfirst(char *mech, void *rock __attribute__((unused)))
     printf("%s --> successful result\n", mech);
 }
 
-void test_serverlast(char *mech, void *rock __attribute__((unused))) 
+void test_serverlast(char *mech, void *rock) 
 {
     sasl_conn_t *sconn, *cconn;
+    tosend_t *tosend = (tosend_t *)rock;
     
     printf("%s --> start\n", mech);
 
     /* Basic crash-tests (none should cause a crash): */
-    if(doauth_serverlast(mech, &sconn, &cconn, &security_props) != SASL_OK) {
+    if(doauth_serverlast(mech, &sconn, &cconn, &security_props,
+			 tosend->client_callbacks) != SASL_OK) {
 	fatal("doauth failed in test_serverlast");
     }
 
@@ -1957,16 +2164,18 @@ void test_serverlast(char *mech, void *rock __attribute__((unused)))
 }
 
 
-void test_noclientfirst_andserverlast(char *mech,
-				      void *rock __attribute__((unused))) 
+void test_noclientfirst_andserverlast(char *mech, void *rock) 
 {
     sasl_conn_t *sconn, *cconn;
+    tosend_t *tosend = (tosend_t *)rock;
     
     printf("%s --> start\n", mech);
 
     /* Basic crash-tests (none should cause a crash): */
     if(doauth_noclientfirst_andserverlast(mech, &sconn, &cconn,
-					  &security_props) != SASL_OK) {
+					  &security_props,
+					  tosend->client_callbacks)
+       != SASL_OK) {
 	fatal("doauth failed in test_noclientfirst_andserverlast");
     }
 
@@ -1983,14 +2192,15 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     const char *txstring = "THIS IS A TEST";
     const char *out, *out2;
     char *tmp;
-    const sasl_security_properties_t *test_props[6] =
+    const sasl_security_properties_t *test_props[7] =
                                           { &security_props,
 					    &force_3des,
+					    &force_rc4_56,
 					    &force_des,
 					    &int_only,
 					    &no_int,
 					    &disable_seclayer };
-    const unsigned num_properties = 6;
+    const unsigned num_properties = 7;
     unsigned i;
     const sasl_ssf_t *this_ssf;
     unsigned outlen = 0, outlen2 = 0, totlen = 0;
@@ -2000,7 +2210,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     for(i=0; i<num_properties; i++) {
         
     /* Basic crash-tests (none should cause a crash): */
-    result = doauth(mech, &sconn, &cconn, test_props[i], 1);
+    result = doauth(mech, &sconn, &cconn, test_props[i], NULL, 1);
     if(result == SASL_NOMECH && test_props[i]->min_ssf > 0) {
 	printf("  Testing SSF: SKIPPED (requested minimum > 0: %d)\n",
 	       test_props[i]->min_ssf);
@@ -2048,7 +2258,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Basic I/O Test */
-    if(doauth(mech, &sconn, &cconn, test_props[i], 0) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i], NULL, 0) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -2065,7 +2275,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Split one block and reassemble */
-    if(doauth(mech, &sconn, &cconn, test_props[i], 0) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i], NULL, 0) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -2103,7 +2313,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Combine 2 blocks */
-    if(doauth(mech, &sconn, &cconn, test_props[i], 0) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i], NULL, 0) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -2140,7 +2350,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Combine 2 blocks with 1 split */
-    if(doauth(mech, &sconn, &cconn, test_props[i], 0) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i], NULL, 0) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -2236,6 +2446,9 @@ void foreach_mechanism(foreach_t *func, void *rock)
 	fatal("sasl_server_new in foreach_mechanism");
     }
 
+    if (sasl_setprop(saslconn, SASL_AUTH_EXTERNAL, authname)!=SASL_OK)
+	fatal("sasl_setprop(SASL_AUTH_EXTERNAL) failed");
+
     result = sasl_listmech(saslconn,
 			   NULL,
 			   "",
@@ -2347,6 +2560,7 @@ void test_rand_corrupt(unsigned steps)
     {
 	tosend.type = rand() % CORRUPT_SIZE;
 	tosend.step = lup % MAX_STEPS;
+	tosend.client_callbacks = NULL;
 
 	printf("RANDOM TEST: (%s in step %d) (%d of %d)\n",corrupt_types[tosend.type],tosend.step,lup+1,steps);
 	foreach_mechanism((foreach_t *) &sendbadsecond,&tosend);
@@ -2361,7 +2575,8 @@ void test_proxypolicy()
 void test_all_corrupt() 
 {
     tosend_t tosend;
-    
+    tosend.client_callbacks = NULL;
+
     /* Start just beyond NOTHING */
     for(tosend.type=1; tosend.type<CORRUPT_SIZE; tosend.type++) {
 	for(tosend.step=0; tosend.step<MAX_STEPS; tosend.step++) {
@@ -2422,9 +2637,10 @@ void create_ids(void)
     
     result = sasl_checkpass(saslconn, username, strlen(username),
 			    password, strlen(password));
-    if (result != SASL_OK)
+    if (result != SASL_OK) {
+	fprintf(stderr, "%s\n", sasl_errdetail(saslconn));
 	fatal("Unable to verify password we just set");
-
+    }
     result = sasl_user_exists(saslconn, "imap", NULL, username);
     if(result != SASL_OK)
 	fatal("sasl_user_exists did not find user");
@@ -2668,7 +2884,7 @@ int main(int argc, char **argv)
     if(mem_stat() != SASL_OK) fatal("memory error");
     printf("ok\n");
 
-    printf("Tests of sasl_server_init()... ");
+    printf("Tests of sasl_{server|client}_init()... ");
     test_init();
     if(mem_stat() != SASL_OK) fatal("memory error");
     printf("ok\n");
@@ -2688,6 +2904,31 @@ int main(int argc, char **argv)
 	
 	tosend.type = NOTHING;
 	tosend.step = 500;
+	tosend.client_callbacks = client_interactions;
+	
+	printf("Testing client-first/no-server-last correctly...\n");
+	foreach_mechanism((foreach_t *) &test_clientfirst,&tosend);
+	if(mem_stat() != SASL_OK) fatal("memory error");
+	printf("Test of client-first/no-server-last...ok\n");
+
+	printf("Testing no-client-first/no-server-last correctly...\n");
+	foreach_mechanism((foreach_t *) &test_noclientfirst, &tosend);
+	if(mem_stat() != SASL_OK) fatal("memory error");
+	printf("Test of no-client-first/no-server-last...ok\n");
+	
+	printf("Testing no-client-first/server-last correctly...\n");
+	foreach_mechanism((foreach_t *) &test_noclientfirst_andserverlast,
+			  &tosend);
+	if(mem_stat() != SASL_OK) fatal("memory error");
+	printf("Test of no-client-first/server-last...ok\n");
+
+	printf("Testing client-first/server-last correctly...\n");
+	foreach_mechanism((foreach_t *) &test_serverlast, &tosend);
+	if(mem_stat() != SASL_OK) fatal("memory error");
+	printf("Test of client-first/server-last...ok\n");
+
+	tosend.client_callbacks = client_callbacks;
+	printf("-=-=-=-=- And now using the callbacks interface -=-=-=-=-\n");
 
 	printf("Testing client-first/no-server-last correctly...\n");
 	foreach_mechanism((foreach_t *) &test_clientfirst,&tosend);
@@ -2700,13 +2941,13 @@ int main(int argc, char **argv)
 	printf("Test of no-client-first/no-server-last...ok\n");
 	
 	printf("Testing no-client-first/server-last correctly...\n");
-	foreach_mechanism((foreach_t *) &test_serverlast, &tosend);
+	foreach_mechanism((foreach_t *) &test_noclientfirst_andserverlast,
+			  &tosend);
 	if(mem_stat() != SASL_OK) fatal("memory error");
 	printf("Test of no-client-first/server-last...ok\n");
 
 	printf("Testing client-first/server-last correctly...\n");
-	foreach_mechanism((foreach_t *) &test_noclientfirst_andserverlast,
-			  &tosend);
+	foreach_mechanism((foreach_t *) &test_serverlast, &tosend);
 	if(mem_stat() != SASL_OK) fatal("memory error");
 	printf("Test of client-first/server-last...ok\n");
     } else {
@@ -2714,6 +2955,7 @@ int main(int argc, char **argv)
 	printf("Testing no-client-first/no-server-last correctly...skipped\n");
 	printf("Testing no-client-first/server-last correctly...skipped\n");
 	printf("Testing client-first/server-last correctly...skipped\n");
+	printf("Above tests with callbacks interface...skipped\n");
     }
     
     /* FIXME: do memory tests below here on the things
